/*
 * Copyright (c) 2015, 2016, 2017 Adrian Siekierka
 *
 * This file is part of Preston.
 *
 * Preston is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Preston is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with Preston.  If not, see <http://www.gnu.org/licenses/>.
 */

package pl.asie.preston.machine;

import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.NetworkManager;
import net.minecraft.network.play.server.SPacketUpdateTileEntity;
import net.minecraft.tileentity.TileEntity;
import net.minecraft.util.EnumFacing;
import net.minecraft.util.ITickable;
import net.minecraftforge.common.capabilities.Capability;
import net.minecraftforge.common.util.Constants;
import net.minecraftforge.energy.CapabilityEnergy;
import net.minecraftforge.items.CapabilityItemHandler;
import net.minecraftforge.items.IItemHandler;
import net.minecraftforge.items.ItemStackHandler;
import pl.asie.preston.PrestonMod;
import pl.asie.preston.api.ICompressorRecipe;
import pl.asie.preston.container.ItemCompressedBlock;
import pl.asie.preston.network.PacketEmitParticlesOfHappiness;
import pl.asie.preston.network.PacketSyncHeadProgress;
import pl.asie.preston.util.*;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.List;

public class TileCompressor extends TileBase implements ITickable {
	public int armProgressClient;
	public EnergySystem currentSystem = EnergySystem.FORGE;
	private final VeryLargeMachineEnergyStorage storage;
	private final ItemStackHandler stackHandler;
	private final IItemHandler viewTop, viewSide, viewBottom;
	private final List<EnergyWrapper> wrapperList;
	private boolean shouldShift;

	public TileCompressor() {
		this.storage = new VeryLargeMachineEnergyStorage() {
			@Override
			public int receiveEnergy(int maxReceive, boolean simulate) {
				int r = super.receiveEnergy(maxReceive, simulate);
				if (!simulate && r > 0) {
					setCurrentSystem(EnergySystem.FORGE);
				}
				return r;
			}
		};

		this.wrapperList = new ArrayList<>();
		for (EnergySystem system : EnergySystem.values()) {
			if (system != EnergySystem.FORGE && system.isEnabled() && system.canFunction()) {
				EnergyWrapper w = system.createWrapper(this::setCurrentSystem, storage);
				if (w != null) {
					wrapperList.add(w);
				}
			}
		}

		this.stackHandler = new ItemStackHandler(10) {
			@Override
			@Nonnull
			public ItemStack insertItem(int slot, @Nonnull ItemStack stack, boolean simulate) {
				if (slot == 9) return stack;
				else return super.insertItem(slot, stack, simulate);
			}

			@Override
			protected void onContentsChanged(int slot) {
				shouldShift = true;
			}
		};


		viewTop = new ItemHandlerProxy(stackHandler, 0, 9, true, false);
		viewSide = new ItemHandlerProxy(stackHandler, 0, 10, true, true) {
			@Override
			@Nonnull
			public ItemStack extractItem(int slot, int amount, boolean simulate) {
				if (slot < 9) return ItemStack.field_190927_a;
				else return super.extractItem(slot, amount, simulate);
			}
		};
		viewBottom = new ItemHandlerProxy(stackHandler, 9, 1, false, true);
	}

	private void setCurrentSystem(EnergySystem system) {
		this.currentSystem = system;
	}

	public static ICompressorRecipe getMatchingRecipe(ItemStack stack) {
		if (!stack.func_190926_b()) {
			for (ICompressorRecipe recipe : PrestonMod.recipes) {
				if (recipe.matchesType(stack)) {
					return recipe;
				}
			}
		}

		return null;
	}

	public VeryLargeMachineEnergyStorage getStorage() {
		return storage;
	}

	public IItemHandler getStackHandler() {
		return stackHandler;
	}

	public void readNBTData(NBTTagCompound compound, boolean isClient) {
		if (compound.func_150297_b("inv", Constants.NBT.TAG_COMPOUND)) {
			stackHandler.deserializeNBT(compound.func_74775_l("inv"));
		}

		if (compound.func_150297_b("energy", Constants.NBT.TAG_COMPOUND)) {
			storage.deserializeNBT(compound.func_74775_l("energy"));
		}

		if (isClient && compound.func_150297_b("p", Constants.NBT.TAG_ANY_NUMERIC)) {
			armProgressClient = compound.func_74762_e("p");
		}

		if (compound.func_150297_b("es", Constants.NBT.TAG_ANY_NUMERIC)) {
			currentSystem = EnergySystem.values()[compound.func_74762_e("es")];
		}
	}

	public NBTTagCompound serializeNBTTooltip() {
		NBTTagCompound compound = new NBTTagCompound();
		compound.func_74774_a("es", (byte) currentSystem.ordinal());
		compound.func_74782_a("energy", storage.serializeNBT());
		return compound;
	}

	public NBTTagCompound writeNBTData(NBTTagCompound compound, boolean isClient) {
		if (!isClient) {
			if (shouldShift) {
				shiftInv();
			}
		}

		compound.func_74782_a("inv", stackHandler.serializeNBT());
		compound.func_74782_a("energy", storage.serializeNBT());
		if (isClient) {
			compound.func_74768_a("p", armProgressClient);
		}
		return compound;
	}

	@Override
	@Nullable
	public boolean hasCapability(Capability<?> capability, @Nullable EnumFacing facing) {
		for (EnergyWrapper wrapper : wrapperList) {
			if (wrapper.hasCapability(capability, facing)) {
				return true;
			}
		}
		return capability == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY || (capability == CapabilityEnergy.ENERGY && EnergySystem.FORGE.isEnabled()) || super.hasCapability(capability, facing);
	}

	@Override
	@Nullable
	public <T> T getCapability(Capability<T> capability, @Nullable EnumFacing facing) {
		for (EnergyWrapper wrapper : wrapperList) {
			T value = wrapper.getCapability(capability, facing);
			if (value != null) {
				return value;
			}
		}

		if (capability == CapabilityItemHandler.ITEM_HANDLER_CAPABILITY) {
			if (facing == null) {
				return CapabilityItemHandler.ITEM_HANDLER_CAPABILITY.cast(stackHandler);
			} else {
				switch (facing) {
					case UP:
						return CapabilityItemHandler.ITEM_HANDLER_CAPABILITY.cast(viewTop);
					case DOWN:
						return CapabilityItemHandler.ITEM_HANDLER_CAPABILITY.cast(viewBottom);
					default:
						return CapabilityItemHandler.ITEM_HANDLER_CAPABILITY.cast(viewSide);
				}
			}
		} else if (capability == CapabilityEnergy.ENERGY && EnergySystem.FORGE.isEnabled()) {
			return CapabilityEnergy.ENERGY.cast(storage);
		} else {
			return super.getCapability(capability, facing);
		}
	}

	private void shiftInv() {
		if (stackHandler.getStackInSlot(0).func_190926_b()) {
			for (int i = 1; i < 9; i++) {
				ItemStack stack = stackHandler.getStackInSlot(i);
				if (!stack.func_190926_b()) {
					int maxCopy = 9 - i;
					for (int j = 0; j < 9; j++) {
						if (j < maxCopy) {
							stackHandler.setStackInSlot(j, stackHandler.getStackInSlot(i + j));
						} else {
							stackHandler.setStackInSlot(j, ItemStack.field_190927_a);
						}
					}
					break;
				}
			}
		}
		shouldShift = false;
	}


	public void calcArmProgressClient(boolean canProcess) {
		int newApc = 0;

		if (canProcess) {
			if (storage.getCurrentMaxEnergy().compareTo(BigInteger.ZERO) > 0) {
				BigInteger result = storage.getBigEnergyStored().multiply(BigInteger.valueOf(127)).divide(storage.getCurrentMaxEnergy());
				if (result.compareTo(BigInteger.valueOf(Integer.MAX_VALUE)) <= 0) {
					newApc = 128 + result.intValueExact();
				}
			}
		}

		if (newApc != armProgressClient) {
			armProgressClient = newApc;
			PrestonMod.packet.sendToWatching(new PacketSyncHeadProgress(this), this);
		}
	}

	@Override
	public void func_73660_a() {
		if (field_145850_b.field_72995_K) {
			return;
		}

		if (shouldShift || func_145831_w().func_82737_E() % 10 == 0) {
			shiftInv();
		}

		ItemStack job = stackHandler.getStackInSlot(0);
		ICompressorRecipe recipe = getMatchingRecipe(job);

		if (recipe == null) {
			calcArmProgressClient(false);
			storage.setCurrentMaxEnergy(BigInteger.ZERO);
			return;
		}

		int availableStacks = job.func_190916_E();

		for (int i = 1; i < 9 && availableStacks < recipe.getRequiredItemCount(); i++) {
			ItemStack stack = stackHandler.getStackInSlot(i);
			if (!stack.func_190926_b() && recipe.canMerge(stack, job)) {
				availableStacks += stack.func_190916_E();
			}
		}

		storage.setCurrentMaxEnergy(recipe.getEnergyUsage(job));
		if (availableStacks < recipe.getRequiredItemCount()) {
			calcArmProgressClient(false);
			return;
		}

		calcArmProgressClient(true);

		if (storage.getBigEnergyStored().compareTo(storage.getCurrentMaxEnergy()) >= 0) {
			// Compress
			ItemStack compressed = recipe.getResult(job);
			if (compressed != null) {
				boolean pushedIn = false;
				ItemStack targetSlot = stackHandler.getStackInSlot(9);
				if (targetSlot.func_190926_b()) {
					stackHandler.setStackInSlot(9, compressed);
					pushedIn = true;
				} else if (PrestonUtils.canMerge(targetSlot, compressed) && (targetSlot.func_190916_E()+compressed.func_190916_E()) <= targetSlot.func_77976_d()) {
					targetSlot.func_190917_f(compressed.func_190916_E());
					pushedIn = true;
				}

				if (pushedIn) {
					recipe.onCraftSuccess(func_145831_w(), func_174877_v(), compressed);

					storage.setBigEnergyStored(storage.getBigEnergyStored().subtract(storage.getCurrentMaxEnergy()));
					// Remove used stacks
					int toRemove = recipe.getRequiredItemCount();
					for (int i = 0; i < 9 && toRemove > 0; i++) {
						ItemStack stack = stackHandler.getStackInSlot(i);
						if (!stack.func_190926_b() && (i == 0 || recipe.canMerge(stack, job))) {
							if (stack.func_190916_E() > toRemove) {
								stack.func_190918_g(toRemove);
								toRemove = 0;
							} else {
								stackHandler.setStackInSlot(i, ItemStack.field_190927_a);
								toRemove -= stack.func_190916_E();
								shouldShift = true;
							}
						}
					}
				}
			}
		}
	}
}
